#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Thu Jan 11 2024 14:12:29 by codeskyblue
"""


from __future__ import annotations

import logging
import os
import shlex

import click
from pydantic import BaseModel
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.cli.cli_common import print_json
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo

from tidevice3.cli.cli_common import cli, gcfg
from tidevice3.utils.common import print_dict_as_table
from tidevice3.utils.download import download_file, is_hyperlink


@cli.group()
def app():
    """app related commands"""
    pass


@app.command("install")
@click.argument("path_or_url")
def app_install(path_or_url: str):
    """install given .ipa"""
    if is_hyperlink(path_or_url):
        ipa_path = download_file(path_or_url)
    elif os.path.isfile(path_or_url):
        ipa_path = path_or_url
    else:
        raise ValueError("local file not found", path_or_url)
    service_provider = gcfg.get_lockdown_client()
    InstallationProxyService(lockdown=service_provider).install_from_local(ipa_path)


@app.command("list")
@click.option("-u", "--user", is_flag=True, help="include user apps")
@click.option("-s", "--system", is_flag=True, help="include system apps")
@click.option("--hidden", is_flag=True, help="include hidden apps")
@click.option("--calculate-sizes/--no-calculate-size", default=False)
def app_list(user: bool, system: bool, hidden: bool, calculate_sizes: bool):
    """list installed apps"""
    service_provider = gcfg.get_lockdown_client()
    app_types = []
    if user:
        app_types.append("User")
    if system:
        app_types.append("System")
    if hidden:
        app_types.append("Hidden")
    app_infos = InstallationProxyService(lockdown=service_provider).get_apps(
        app_types, calculate_sizes=calculate_sizes
    )
    print_dict_as_table(app_infos.values(), ["CFBundleIdentifier", "CFBundleDisplayName", "CFBundleVersion", "CFBundleShortVersionString"])


# @app.command()
# @click.argument("bundle_identifier")
# def info(bundle_identifier: str):
#     """show app info"""
#     pass


@app.command("uninstall")
@click.argument("bundle_identifier")
def app_uninstall(bundle_identifier: str):
    """uninstall application"""
    service_provider = gcfg.get_lockdown_client()
    InstallationProxyService(lockdown=service_provider).uninstall(bundle_identifier)


@app.command("launch")
@click.argument("arguments", type=click.STRING)
@click.option("--kill-existing/--no-kill-existing", default=True, help="Whether to kill an existing instance of this process")
@click.option("--suspended", is_flag=True, help="Same as WaitForDebugger")
@click.option("--env", multiple=True, type=click.Tuple((str, str)), help="Environment variables to pass to process given as a list of key value")
@click.option("--stream", is_flag=True)
def app_launch(arguments: str, kill_existing: bool, suspended: bool, env: tuple, stream: bool):
    """launch application"""
    service_provider = gcfg.get_service_provider()
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        process_control = ProcessControl(dvt)
        parsed_arguments = shlex.split(arguments)
        pid = process_control.launch(
            bundle_id=parsed_arguments[0],
            arguments=parsed_arguments[1:],
            kill_existing=kill_existing,
            start_suspended=suspended,
            environment=dict(env),
        )
        print(f"Process launched with pid {pid}")
        while stream:
            for output_received in process_control:
                logging.getLogger(f"PID:{output_received.pid}").info(
                    output_received.message.strip()
                )


@app.command("kill")
@click.argument("pid", type=click.INT)
def app_kill(pid: int):
    """kill application"""
    service_provider = gcfg.get_service_provider()
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        process_control = ProcessControl(dvt)
        process_control.kill(pid)


@app.command("ps")
@click.option('--color/--no-color', default=True)
def app_ps(color: bool):
    """list running processes"""
    service_provider = gcfg.get_service_provider()
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        processes = DeviceInfo(dvt).proclist()
        for process in processes:
            if 'startDate' in process:
                process['startDate'] = str(process['startDate'])
        print_json(processes, colored=color)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Jan 09 2024 19:09:47 by codeskyblue
"""

from __future__ import annotations

from typing import Optional

import click
from pydantic import BaseModel
from pymobiledevice3.services.diagnostics import DiagnosticsService

from tidevice3.cli.cli_common import cli, gcfg


@cli.command()
def reboot():
    """reboot device"""
    lockdown_client = gcfg.get_lockdown_client()
    with DiagnosticsService(lockdown_client) as diagnostics:
        diagnostics.restart()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 18:10:32 by codeskyblue
"""
from __future__ import annotations
import collections

import click
from dataclasses import dataclass, field
from pymobiledevice3.cli.cli_common import USBMUX_OPTION_HELP
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
import requests

from tidevice3.exceptions import FatalError


DEFAULT_TIMEOUT = 60

@dataclass
class GlobalConfig:
    color: bool = True
    usbmux_address: str | None = None
    _udid: str | None = None
    _product_version: str | None = field(default=None, init=False)

    def _update_property(self):
        """ should be called after property changed is set """
        with self.get_lockdown_client() as lockdown_client:
            self._product_version = lockdown_client.product_version
            self._udid = lockdown_client.udid
    
    @property
    def udid(self) -> str:
        if self._udid:
            return self._udid
        self._update_property()
        return self._udid
    
    @property
    def product_version(self) -> str:
        if self._product_version:
            return self._product_version
        self._update_property()
        return self._product_version

    @property
    def major_version(self) -> int:
        return int(self.product_version.split(".")[0])
    
    def get_lockdown_client(self) -> LockdownClient:
        return create_using_usbmux(serial=self._udid, usbmux_address=self.usbmux_address)
    
    def get_service_provider(self) -> LockdownServiceProvider:
        if self.major_version >= 17:
            return self._get_remote_service_discovery_service()
        return self.get_lockdown_client()
    
    def _get_remote_service_discovery_service(self) -> RemoteServiceDiscoveryService:
        try:
            resp = requests.get("http://localhost:5555", timeout=DEFAULT_TIMEOUT)
            tunnels = resp.json()
            ipv6_address = tunnels.get(self.udid)
            if ipv6_address is None:
                raise FatalError("tunneld not ready for device", self.udid)
            rsd = RemoteServiceDiscoveryService(ipv6_address)
            rsd.connect()
            return rsd
        except requests.RequestException:
            raise FatalError("Please run `sudo t3 tunneld` first")
        except (TimeoutError, ConnectionError):
            raise FatalError("RemoteServiceDiscoveryService connect failed")


gcfg: GlobalConfig = GlobalConfig()


def get_udid() -> str:
    if gcfg.udid:
        return gcfg.udid
    raise RuntimeError("udid not set")


class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, *args, **attrs):
        super(OrderedGroup, self).__init__(name, commands, *args, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands


@click.group(cls=OrderedGroup, context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-u", "--udid", default=None, help="udid of device")
@click.option("usbmux_address", "--usbmux", help=USBMUX_OPTION_HELP)
@click.option("--color/--no-color", default=True)
def cli(udid: str, usbmux_address: str, color: bool):
    gcfg._udid = udid
    gcfg.usbmux_address = usbmux_address
    gcfg.color = color
    

CLI_GROUPS = ["list", "developer", "install", "uninstall", "screenshot", "fsync", "app", "reboot", "tunneld", "exec"]
for group in CLI_GROUPS:
    __import__(f"tidevice3.cli.{group}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 18:10:01 by codeskyblue
"""
from __future__ import annotations

from typing import Optional

import click
from pydantic import BaseModel
from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import print_json
from pymobiledevice3.lockdown import create_using_usbmux, usbmux

from tidevice3.cli.cli_common import cli, gcfg
from tidevice3.utils.common import print_dict_as_table


@cli.command(name="list")
@click.option("-u", "--usb", is_flag=True, help="show only usb devices")
@click.option("-n", "--network", is_flag=True, help="show only network devices")
@click.option("--json", is_flag=True, help="output as json format")
def cli_list(usb: bool, network: bool, json: bool):
    """list connected devices"""
    devices = list_devices(usb, network, gcfg.usbmux_address)
    if json:
        print_json([d.model_dump() for d in devices], colored=gcfg.color)
    else:
        headers = ["Identifier", "DeviceName", "ProductType", "ProductVersion", "ConnectionType"]
        print_dict_as_table([d.model_dump() for d in devices], headers)


class DeviceShortInfo(BaseModel):
    BuildVersion: str
    ConnectionType: Optional[str]
    DeviceClass: str
    DeviceName: str
    Identifier: str
    ProductType: str
    ProductVersion: str


def list_devices(
    usb: bool, network: bool, usbmux_address: Optional[str] = None
) -> list[DeviceShortInfo]:
    connected_devices = []
    for device in usbmux.list_devices(usbmux_address=usbmux_address):
        udid = device.serial

        if usb and not device.is_usb:
            continue

        if network and not device.is_network:
            continue

        lockdown = create_using_usbmux(
            udid,
            autopair=False,
            connection_type=device.connection_type,
            usbmux_address=usbmux_address,
        )
        info = DeviceShortInfo.model_validate(lockdown.short_info)
        connected_devices.append(info)
    return connected_devices



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Jan 08 2024 14:50:21 by codeskyblue
"""

import io
import logging
import typing

import click
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.screenshot import ScreenshotService
from pymobiledevice3.services.dvt.instruments.screenshot import Screenshot

from tidevice3.cli.cli_common import cli, gcfg

logger = logging.getLogger(__name__)

from PIL import Image


def byte2image(data: bytes) -> Image:
    return Image.open(io.BytesIO(data)).convert("RGB")


def screenshot(lockdown_client: LockdownClient) -> Image.Image:
    service_provider = gcfg.get_service_provider()
    if gcfg.major_version >= 17:
        with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
            png_data = Screenshot(dvt).get_screenshot()
    else:
        png_data = ScreenshotService(lockdown_client).take_screenshot()
    return byte2image(png_data)
    

@cli.command("screenshot")
@click.argument("out", type=click.File("wb"))
def cli_screenshot(out: typing.BinaryIO):
    """get device screenshot"""
    lockdown_client = gcfg.get_lockdown_client()
    screenshot(lockdown_client).save(out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Jan 05 2024 20:03:54 by codeskyblue
"""

import logging
import os

import click
from pymobiledevice3.services.installation_proxy import \
    InstallationProxyService

from tidevice3.cli.cli_common import cli, gcfg
from tidevice3.utils.download import download_file, is_hyperlink

logger = logging.getLogger(__name__)


@cli.command()
@click.argument('path_or_url')
def install(path_or_url: str):
    """ install given .ipa """
    logger.warning("deprecated, use `t3 app install` instead")
    if is_hyperlink(path_or_url):
        ipa_path = download_file(path_or_url)
    elif os.path.isfile(path_or_url):
        ipa_path = path_or_url
    else:
        raise ValueError('local file not found', path_or_url)
    service_provider = gcfg.get_lockdown_client()
    InstallationProxyService(lockdown=service_provider).install_from_local(ipa_path)
